/* 
    Copyright (C) 2004  Mika Raento - Renaud Petit

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


    email: mraento@cs.helsinki.fi - petit@cs.helsinki.fi 
*/


//CBlackBoardServerSession.cpp

#include "BlackBoardServerSession.h"
#include "BlackBoard_cs.h"
#include <e32svr.h>
#include <flogger.h>
#include <symbian_auto_ptr.h>
#include "bberrors.h"

bool ClientAlive(TInt ThreadId)
{
	CALLSTACKITEM(_L("CActiveScheduler::Stop"));


	if (ThreadId==0) return false;

	RThread c; bool ret = false;
	if (c.Open(ThreadId) != KErrNone) return false;
	if (c.ExitType() == EExitPending) ret=true;
	c.Close();
	return ret;
}

class CIdleCallBack : public CActive {
private:
	TBool iEnabled;
	MBack& iBack;
public:

	static CIdleCallBack* NewL(MBack& aBack) {
		auto_ptr<CIdleCallBack> ret(new (ELeave) CIdleCallBack(aBack));
		ret->ConstructL();
		return ret.release();
	}

	CIdleCallBack(MBack& aBack) : CActive(EPriorityIdle), iEnabled(EFalse), iBack(aBack) { }
	void ConstructL() {
		CActiveScheduler::Add(this);
	}
	void RunL() { if(iEnabled) iBack.Back(); }
	void DoCancel() { iEnabled=EFalse; }
	void Reset() { iEnabled=EFalse; }
	void Trigger() { 
		iEnabled=ETrue;
		if (!IsActive()) { 
			TRequestStatus* s=&iStatus;
			User::RequestComplete(s, KErrNone);
			SetActive();
		}
	}
	~CIdleCallBack() { Cancel(); }
};

CBlackBoardServerSession::CBlackBoardServerSession(RThread& aClient, CBlackBoardServer& aServer) : CSession(aClient), iServer(aServer)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::CBlackBoardServerSession"));


    
}

CBlackBoardServerSession::~CBlackBoardServerSession()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::~CBlackBoardServerSession"));


	delete iIdle;

	//FIXME: what happens if server is shutting down?
	iServer.DeleteAllNotificationsL(this);

	TInt i;
	for (i=0; i<KPriorityCount; i++) {
		delete iNotifications[i];
	}

    	iServer.DecrementSessions();
}

void CBlackBoardServerSession::CompleteMessage(TMessageIdxs aWhich, TInt Code)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::CompleteMessage"));


	if (ClientAlive(iMessageThreadId[aWhich]) && iMessage[aWhich]!=RMessage()) {
		iMessage[aWhich].Complete(Code);
	}
	if (aWhich==ENotifyMsgIndex) iWaitingForNotify=EFalse;

	SetMessage(aWhich, RMessage());
}

void CBlackBoardServerSession::ServiceL(const RMessage& aMessage)
{
	GetContext()->ResetCallStack();

	CALLSTACKITEM(_L("CBlackBoardServerSession::ServiceL"));


	TMessageIdxs msgidx;

	msgidx=EOtherMsgIndex;
	if (aMessage.Function()!=ECancelOther && aMessage.Function()!=ECancelNotify) {
		// allow other calls while waiting for notify
		if (aMessage.Function()==ENotifyOnChange) {
			msgidx=ENotifyMsgIndex;	
		}
		SetMessage(msgidx, aMessage);
		ReadFlags(msgidx);
	}
	
	switch (aMessage.Function())
	{
		case EGetByTuple:
			GetByTupleL();
			break;
		case EGetByComponent:
			GetByComponentL();
			break;
		case EPut:
			PutL();
			break;
		case EDeleteByTuple:
			DeleteByTupleL();
			break;
		case EDeleteById:
			DeleteByIdL();
			break;
		case EAddNotifyByTupleFilter:
			AddTupleFilterL();
			break;
		case EDeleteNotifyByTupleFilter:
			DeleteTupleFilterL();
			break;
		case EAddNotifyByComponentFilter:
			AddComponentFilterL();
			break;
		case EDeleteNotifyByComponentFilter:
			DeleteComponentFilterL();
			break;
		case ENotifyOnChange:
			if (!SendWaitingL()) iWaitingForNotify=ETrue;
			break;
		case ETerminateBlackBoardServer:
			TerminateServer();
			break;
		case ECancelOther:
			// hmm. Since all of these messages get
			// Complete()d when they come in, there isn't
			// a pending message, which makes this is a NOP
			// it'll do something, tho, if we change
			// some message to be really async
			CompleteMessage(EOtherMsgIndex, KErrCancel);
			// but this should be correct even if the
			// above does nothing
			aMessage.Complete(KErrNone);
			break;
		case ECancelNotify:
			iWaitingForNotify=EFalse;
			CompleteMessage(ENotifyMsgIndex, KErrCancel);
			aMessage.Complete(KErrNone);
			break;
		default :
            	PanicClient(EBadRequest);
	}

}

void CBlackBoardServerSession::PanicClient(TInt aPanic) const
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::PanicClient"));


	Panic(KBlackBoardServer, aPanic) ; // Note: this panics the client thread, not server
}

CBlackBoardServerSession* CBlackBoardServerSession::NewL(RThread& aClient, CBlackBoardServer& aServer)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::NewL"));


	CBlackBoardServerSession* self = CBlackBoardServerSession::NewLC(aClient, aServer) ;
    	CleanupStack::Pop(self) ;
    	return self ;
}

CBlackBoardServerSession* CBlackBoardServerSession::NewLC(RThread& aClient, CBlackBoardServer& aServer)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::NewLC"));


	CBlackBoardServerSession* self = new (ELeave) CBlackBoardServerSession(aClient, aServer);
    	CleanupStack::PushL(self) ;
    	self->ConstructL();
    	return self ;
}

void CBlackBoardServerSession::ConstructL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ConstructL"));

	iServer.IncrementSessions();

	iIdle=CIdleCallBack::NewL(*this);

	TInt i;
	for (i=0; i<KPriorityCount; i++) {
		iNotifications[i]=CList<TUint>::NewL();
	}
}

//------------------------------------------------------------------------

void CBlackBoardServerSession::SetMessage(TMessageIdxs aWhich, const RMessage& aMessage) 
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::SetMessage"));


	if (aMessage==RMessage()) {
		iMessageThreadId[aWhich]=0;
	} else {
		iMessageThreadId[aWhich]=aMessage.Client().Id();
	}
	iMessage[aWhich]=aMessage;
}

void CBlackBoardServerSession::ReadTupleAndSubL(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadTupleAndSubL"));

	if ( iFlags & KArgFlagNoSubName ) {
		iFullArgs.iSubName.Zero();
		TPckg<TTupleName> p(iFullArgs.iTupleName);
		iMessage[aWhich].ReadL( iMessage[aWhich].Ptr1(), p );
	} else {
		TTupleArgs t;
		TPckg<TTupleArgs> p(t);
		iMessage[aWhich].ReadL( iMessage[aWhich].Ptr1(), p );
		iFullArgs.iTupleName=t.iTupleName;
		iFullArgs.iSubName=t.iSubName;
	}
}

void CBlackBoardServerSession::ReadComponentL(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadComponentL"));

	iFullArgs.iComponentName.iModule.iUid=iMessage[aWhich].Int1();
	iFullArgs.iComponentName.iId=iMessage[aWhich].Int2();
}


void CBlackBoardServerSession::ReadId(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadId"));

	iFullArgs.iId=(TUint)iMessage[aWhich].Ptr1();
}

void CBlackBoardServerSession::GetByTupleL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::GetByTupleL"));

	ReadTupleAndSubL(EOtherMsgIndex);
	RADbColReadStream rs;
	iServer.GetL(iFullArgs.iTupleName, iFullArgs.iSubName, iFullArgs.iId, iFullArgs.iComponentName, rs, iSize);
	WriteDataL(rs, EOtherMsgIndex);
}

void CBlackBoardServerSession::ReadFlags(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadFlags"));

	iFlags=(TUint)iMessage[aWhich].Ptr0();
}

void CBlackBoardServerSession::GetByComponentL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::GetByComponentL"));

	ReadComponentL(EOtherMsgIndex);
	RADbColReadStream rs;
	iServer.GetL(iFullArgs.iComponentName, iFullArgs.iId, iFullArgs.iTupleName, iFullArgs.iSubName, rs, iSize);
	WriteDataL(rs, EOtherMsgIndex);
}

void CBlackBoardServerSession::ReadFullArgsL(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadFullArgsL"));

	TPckg<TFullArgs> p(iFullArgs);
	p.SetLength(0);
	iMessage[aWhich].ReadL( iMessage[aWhich].Ptr1(), p );
}

HBufC8* CBlackBoardServerSession::ReadDataL(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadDataL"));

	TInt len;
	len=iMessage[aWhich].Client().GetDesLength(iMessage[aWhich].Ptr3());
	if (len<0) User::Leave(len);
	auto_ptr<HBufC8> ret(HBufC8::NewL(len));
	TPtr8 p=ret->Des();
	iMessage[aWhich].ReadL( iMessage[aWhich].Ptr3(), p );
	return ret.release();
}

void CBlackBoardServerSession::PutL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::PutL"));

	ReadFullArgsL(EOtherMsgIndex);
	TBool Replace=EFalse;
	if (iFlags & KArgFlagReplace) Replace=ETrue;

	auto_ptr<HBufC8> b(ReadDataL(EOtherMsgIndex));

	// FIXME: return warning if PutL returns an error code
	// TInt ret=
	TBool persist=ETrue;
	if (iFlags & KArgFlagDontPersist) persist=EFalse;

	if (iFlags & KArgFlagDoNotNotifySender)	iInPut=ETrue;

	TRAPD(err, iServer.PutL( iFullArgs.iTupleName, iFullArgs.iSubName, iFullArgs.iComponentName,
		*b, iFullArgs.iPriority, Replace, iFullArgs.iId, persist ));
	iInPut=EFalse;
	User::LeaveIfError(err);

	if (iMessage[EOtherMsgIndex].Ptr2()) {
		TPckg<TUint> p(iFullArgs.iId);
		iMessage[EOtherMsgIndex].WriteL( iMessage[EOtherMsgIndex].Ptr2(), p );
	}
	CompleteMessage(EOtherMsgIndex, KErrNone);
}

void CBlackBoardServerSession::DeleteByTupleL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::DeleteByTupleL"));

	ReadTupleAndSubL(EOtherMsgIndex);
	iServer.DeleteL(iFullArgs.iTupleName, iFullArgs.iSubName);
	CompleteMessage(EOtherMsgIndex, KErrNone);
}

void CBlackBoardServerSession::DeleteByIdL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::DeleteByIdL"));

	ReadId(EOtherMsgIndex);
	iServer.DeleteL(iFullArgs.iId);
	CompleteMessage(EOtherMsgIndex, KErrNone);
}

void CBlackBoardServerSession::ReadPriorityL(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadPriorityL"));

	iFullArgs.iPriority=(TBBPriority)iMessage[aWhich].Int3();
}

void CBlackBoardServerSession::ReadTupleL(TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::ReadTupleL"));

	iFullArgs.iTupleName.iModule.iUid=iMessage[aWhich].Int1();
	iFullArgs.iTupleName.iId=iMessage[aWhich].Int2();
}

void CBlackBoardServerSession::AddTupleFilterL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::AddTupleFilterL"));

	ReadTupleL(EOtherMsgIndex);
	ReadPriorityL(EOtherMsgIndex);
	TBool GetExisting=EFalse;
	if (iFlags & KArgFlagGetExisting) GetExisting=ETrue;

	iServer.AddNotificationL(this, iFullArgs.iTupleName,
		GetExisting, iFullArgs.iPriority);

	CompleteMessage(EOtherMsgIndex, KErrNone);
}

void CBlackBoardServerSession::DeleteTupleFilterL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::DeleteTupleFilterL"));

	ReadTupleAndSubL(EOtherMsgIndex);
	iServer.DeleteNotificationL(this, iFullArgs.iTupleName);
	CompleteMessage(EOtherMsgIndex, KErrNone);
}

void CBlackBoardServerSession::AddComponentFilterL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::AddComponentFilterL"));

	ReadComponentL(EOtherMsgIndex);
	ReadPriorityL(EOtherMsgIndex);
	TBool GetExisting=EFalse;
	if (iFlags & KArgFlagGetExisting) GetExisting=ETrue;
	iServer.AddNotificationL(this, iFullArgs.iComponentName,
		GetExisting, iFullArgs.iPriority);
	CompleteMessage(EOtherMsgIndex, KErrNone);
}

void CBlackBoardServerSession::DeleteComponentFilterL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::DeleteComponentFilterL"));

	ReadComponentL(EOtherMsgIndex);
	iServer.DeleteNotificationL(this, iFullArgs.iComponentName);
	CompleteMessage(EOtherMsgIndex, KErrNone);
}

#pragma warning(disable: 4706)
TBool CBlackBoardServerSession::SendWaitingL()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::SendWaitingL"));

	TInt p;
	for (p=EBBPriorityOOB; p>=EBBPriorityLow; p--) {
		TUint id;
		if ( (id=iNotifications[p]->Top()) ) {
			RADbColReadStream rs;
			iServer.GetL(id, iFullArgs.iTupleName, iFullArgs.iSubName, 
				iFullArgs.iComponentName, rs, iSize);
			iFullArgs.iId=id;
			WriteDataL(rs, ENotifyMsgIndex);
			iNotifications[p]->Pop();
			return ETrue;
		}
	}
	return EFalse;
}

void CBlackBoardServerSession::TerminateServer()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::TerminateServer"));

	CompleteMessage(EOtherMsgIndex, KErrNone);
	iServer.TerminateServer();
}

void CBlackBoardServerSession::NotifyDirectL(TUint aId, TBBPriority aPriority,
		const TTupleName& aTupleName, const TDesC& aSubName, 
		const TComponentName& aComponent,
		const TDesC8& aSerializedData)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::NotifyDirectL"));

	iFullArgs.iId=aId;
	iFullArgs.iTupleName=aTupleName;
	iFullArgs.iSubName=aSubName;
	iFullArgs.iComponentName=aComponent;
	iFullArgs.iPriority=aPriority;

	WriteDataL(aSerializedData, ENotifyMsgIndex);
}

void CBlackBoardServerSession::NotifyL(TUint aId, TBBPriority aPriority,
			const TTupleName& aTupleName, const TDesC& aSubName, 
			const TComponentName& aComponent,
			const TDesC8& aSerializedData)

{
	CALLSTACKITEM(_L("CBlackBoardServerSession::NotifyL"));

	if (iInPut) return;

	TInt err=-1;
	if ( (!aId || aPriority >= KNotifyDirectLimit) && iWaitingForNotify) {
		TRAPD(err, NotifyDirectL(aId, aPriority, aTupleName, aSubName,
			aComponent, aSerializedData));
	} 
	if (err!=KErrNone) {
		if (aId) iNotifications[aPriority]->AppendL(aId);
		if (err==KClientBufferTooSmall) {
			CompleteMessage(ENotifyMsgIndex, KClientBufferTooSmall);
			iWaitingForNotify=EFalse;
		} else if (aId==0) {
			CompleteMessage(ENotifyMsgIndex, err);
		}
		if (aId && iWaitingForNotify) iIdle->Trigger();
	}
}

void CBlackBoardServerSession::NotifyL(TUint aId, TBBPriority aPriority)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::NotifyL"));

#ifdef __WINS__
	TBuf<100> msg=_L("NOTIFYID: ");
	msg.AppendNum(aId);
#endif
	if (iInPut) return;

	iNotifications[aPriority]->AppendL(aId);
	if (iWaitingForNotify) iIdle->Trigger();
}

void CBlackBoardServerSession::Back()
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::Back"));

	if (!iWaitingForNotify) return;
	TRAPD(err, SendWaitingL());
	if (err!=KErrNone) CompleteMessage(ENotifyMsgIndex, err);
	iWaitingForNotify=EFalse;
}

void CBlackBoardServerSession::WriteDataL(const TDesC8& data, TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::WriteDataL::des"));

	if ( (iMessage[aWhich]) ==RMessage()) 
		User::Leave(KErrGeneral);

	TPckg<TFullArgs> meta(iFullArgs);
	iMessage[aWhich].WriteL( iMessage[aWhich].Ptr2(), meta ); 
	iMessage[aWhich].WriteL( iMessage[aWhich].Ptr3(), data );

	CompleteMessage(aWhich, KErrNone);
}

void CBlackBoardServerSession::WriteDataL(RADbColReadStream& rs, TMessageIdxs aWhich)
{
	CALLSTACKITEM(_L("CBlackBoardServerSession::WriteDataL::rs"));

	if ( (iMessage[aWhich])==RMessage()) 
		User::Leave(KErrGeneral);

	TUint len;
	len=iMessage[aWhich].Client().GetDesMaxLength(iMessage[aWhich].Ptr3());
	if (len < iSize ) User::Leave(KClientBufferTooSmall);

	auto_ptr<HBufC8> b(HBufC8::NewL(iSize));
	TPtr8 p=b->Des();
	if (iSize>0) {
		rs.ReadL(p, iSize);
	}

	WriteDataL(p, aWhich);
}